;implementation of scheme catalogs
; assumes srfi 1,99 
;creates catalog->libraries name&version library->versions serialize-library

;types
(define $library (make-rtd 'library '#((immutable name) (immutable versions) (immutable properties))))
(define $version (make-rtd 'version '#((immutable number) (immutable requires) (immutable uri) (immutable properties))))
(define make-library (rtd-constructor $library))
(define make-version (rtd-constructor $version))
(define library? (rtd-predicate $library))
(define version? (rtd-predicate $version))
(define library-name (rtd-accessor $library 'name))
(define library-versions (rtd-accessor $library 'versions))
(define library-properties (rtd-accessor $library 'properties))
(define version-number (rtd-accessor $version 'number))
(define version-requires (rtd-accessor $version 'requires))
(define version-uri (rtd-accessor $version 'uri))
(define version-properties (rtd-accessor $version 'properties))

(define (version-property version name . default)
   (let ((default (if (null? default) #f (car default)))
	 (result (assoc name (version-properties version))))
	   (cond
	      ((and result (null? (cddr result))) (cadr result))
	      (result (cdr result))
	      (else default))))   

 
(define (name&version name)
 (cond
  ((and (symbol? (car name)) (pair? (cdr name)))
   (call-with-values 
    (lambda () (name&version (cdr name)))
    (lambda (n v) (values (cons (car name) n) v))))
  ((and (symbol? (car name)) (null? (cdr name))) (values name '()))
  ((pair? (car name)) (values '() (car name)))))
	  
;parser

	(define catalog->libraries
	 (let ()
	  (define (find proc lst default)
	   (call/cc (lambda (return)
		     (for-each (lambda (f) (if (proc f) (return f))) lst)
		     (return default))))
	  (define (element-ref name lst) 
	   (let ((result (find (lambda (f) (and (pair? f) (eq? (car f) name))) lst '())))
	    (if (pair? result) (cdr result) result)))

	  (define (form->version form base)
	   (cond
	    ((string? form) (form->version (list '() form)))
	    (else
	     (let ((properties (element-ref '@ (cdr form))) 
		   (requires (element-ref 'requires (cdr form)))
		   (version (car form))
		   (uri (find string? (cdr form) #f)))
	      (make-version version requires (if uri (path-expand uri base) "") properties) 
	     ))))

	  (define (read-catalog path)
	   (with-exception-catcher 
			(lambda (e) '())
		   	(lambda () (with-input-from-file path read)))) 


	(define (main base-uri body)
	 (define (form->catalog form)
	  (if (string? (cadr form)) 
	   (catalog->libraries (cadr form) 
		(cons (read-catalog (cadr form)) (cddr form))) 
	   (catalog->libraries base-uri (cdr form))))
	 (define (form->library form) 
	  (let ((name (cadr form))
		(versions (element-ref 'versions (cddr form)))
		(properties (element-ref '@ (cddr form)))
		(base (find string? (cddr form) base-uri)))
	   (make-library name (map (lambda (f) (form->version f base)) versions) properties)))
	 (let ((result '()))
	  ;generate libraries and flatten any internal lists
	  (for-each (lambda (x) 
		     (cond 
		      ((pair? x) (set! result (append x result)))
		      ((null? x) #f)
		      (else (set! result (cons x result)))))
	   (if (pair? body)
		(map (lambda (f)
		 (case (car f)
		  ((catalog) (form->catalog f))
		  ((library) (form->library f))
		  ((@) '())
		  (else '())))
	    		(cdr body))
		'()))
	  (reverse result)))

	(lambda (base-uri . body)
	 (if (null? body)
	  (main base-uri (read-catalog base-uri))
	  (main base-uri (car body))))))			
;remove this
#;	(define library->versions 
	 (let ()


	(lambda (name)
	 (call-with-values (lambda () (name&version name))
	  (lambda (name version-reference)
	   (let ((library (hash-table-ref CATALOG name #f))
		 (version? (version-predicate version-reference))
		 (result '()))
	    (for-each (lambda (version)
		       (if (version? (version-number version)) (append (list version) result)))
	     (or (and library (library-versions library)) '()))
	    result))))
	))


(define (serialize-library l . base)
	(define (last-index-of s c)
	    (let loop ((i (string-length s)))
		(if (< i 0) #f
		(if (char=? c (string-ref s i)) i (loop (- i 1)))))) 
	(define (relativize path base)
		(let ((l1 (string-length path))
		      (l2 (string-length base)))
		(cond 
		   ((zero? l2) path)
		   ((char=? (string-ref path 0) "/") path)
		   (else (string-append (substring base 0 (last-index-of base #\/)) path)))))

	(define (serialize-version v)
		`(,(version-number v) 
		  ,@(if (null? (version-requires v)) 
			'() 
			`((requires ,@(version-requires v))))
		  ,(if (null? base) (version-uri v) (relativize (version-uri v) (car base)))
		  ,@(if (null? (version-properties v))
			'()
			`((@ ,@(version-properties v))))))
    `(library 
	,(library-name l) 
	,@base
	,@(if (null? (library-properties l)) '() `((@ ,@(library-properties l))))
	(versions ,@(map serialize-version (library-versions l)))))

(define (satisfies-requirements? requirement features)
      (define (satisfied? requirement)
	(cond 
	  ((null? requirement) #t)
	  ((pair? requirement)
	    (case (car requirement)
	      ((and)
	       (all-satisfied? (cdr requirement)))
	      ((or)
	       (any-satisfied? (cdr requirement)))
	      ((not)
	       (not (satisfied? (cadr requirement))))
		;otherwise it is the same as and
	      (else (all-satisfied? requirement))))
		;we assume that requirement is a single symbol
	    (else (memq requirement features))))
      
      ; True if every requirement in LIST is satisfied.

      (define (all-satisfied? list)
	(if (null? list)
	    #t
	    (and (satisfied? (car list))
		 (all-satisfied? (cdr list)))))
      
      ; True if any requirement in LIST is satisfied.

      (define (any-satisfied? list)
	(if (null? list)
	    #f
	    (or (satisfied? (car list))
		(any-satisfied? (cdr list)))))

	
	(satisfied? requirement))

;returns the most specific version based on name version-reference and features
	(define (find-version catalog name version-reference features)
	 (define (find-library catalog name)
	  (hash-table-ref/default catalog name #f))
	 (let ((library (find-library catalog name)))
	  (if library
	   ;return the first version that satisfies the version reference and requires clause
	   (car (filter (lambda (v) 
			(and (version-reference v) (satisfies-requirements? (version-requires v) features))) (library-versions library)))
		#f)))
		
	       
	
