(library (rnrs conditions (6))
	 (export &condition condition simple-conditions condition? condition-predicate condition-accessor
		 define-condition-type
		 &message make-message-condition message-condition? condition-message
		 &warning make-warning warning?
		 &serious make-serious-condition serious-condition?
		 &error make-error error?
		 &violation make-violation violation?
		 &assertion make-assertion-violation assertion-violation?
		 &irritants make-irritants-condition irritants-condition? condition-irritants
		 &who make-who-condition who-condition? condition-who
		 &non-continuable make-non-continuable-violation non-continuable-violation?
		 &implementation-restriction make-implementation-restriction-violation implementation-restriction-violation?
		 &lexical make-lexical-violation lexical-violation?
		 &syntax make-syntax-violation syntax-violation-form syntax-violation-subform
		 &undefined make-undefined-violation undefined-violation?)
	 (import
	   (for (core primitives) expand run)
	   (for (core syntax-rules) expand run)
	   (for (core derived) expand run)
	   (core let)
	   (for (only (core records procedural) 
		      make-rtd rtd-constructor rtd-predicate rtd-accessor) expand run)
	   (primitives pair? null? car cdr list
		       &condition &serious &violation &assertion &who &message &irritants &syntax
		       condition simple-conditions make-assertion-violation make-syntax-violation
		       make-who-condition make-message-condition make-irritants-condition))

	 (define-syntax define-condition-type
	   (syntax-rules ()
		((_ condition-type supertype constructor predicate (field accessor) ...)
		 (begin
		   (define condition-type (make-rtd 'condition-type '#((immutable field) ...) supertype))
		   (define constructor (rtd-constructor condition-type))
		   (define predicate (condition-predicate condition-type))
		   (define accessor (condition-accessor condition-type 
							(rtd-accessor condition-type 'field))) ...))))
	 
	 (define condition? (rtd-predicate &condition))
	   
	 (define (condition-predicate rtd)
	   (define predicate (rtd-predicate rtd))
	   (lambda (x)
	     (let recur ((conditions (simple-conditions x)))
	       (if (null? conditions) #f
		 (if (predicate (car conditions)) #t
		   (recur (cdr conditions)))))))

	 ;should the accessor throw an error if the condition is not the proper type?
	 (define (condition-accessor rtd proc)
	   (define predicate (rtd-predicate rtd))
	   (lambda (x)
	     (let recur ((conditions (simple-conditions x)))
	       (if (pair? conditions)
		 (if (predicate (car conditions))
		   (proc (car conditions))
		   (recur (cdr conditions)))))))

	 ;standard conditions
	 (define-condition-type &warning &condition make-warning warning?)
	 (define make-serious-condition (rtd-constructor &serious))
	 (define serious-condition? (condition-predicate &serious))
	 (define-condition-type &error &serious make-error error?)
	 (define make-violation (rtd-constructor &violation))
	 (define violation? (condition-predicate &violation))
	 (define assertion-violation? (condition-predicate &assertion))
	 (define-condition-type &lexical &violation make-lexical-violation lexical-violation?)
	 (define-condition-type &undefined &violation make-undefined-violation undefined-violation?)
	 (define who-condition? (condition-predicate &who))
	 (define condition-who (condition-accessor &who (rtd-accessor &who 'who)))
	 (define message-condition? (condition-predicate &message))
	 (define condition-message (condition-accessor &message (rtd-accessor &message 'message)))
	 (define irritants-condition? (condition-predicate &irritants))
	 (define condition-irritants (condition-accessor &irritants (rtd-accessor &irritants 'irritants)))

	 (define-condition-type &non-continuable &violation 
				make-non-continuable-violation non-continuable-violation?)
	 (define-condition-type &implementation-restriction &violation 
				make-implementation-restriction-violation 
				implementation-restriction-violation?)
	 (define syntax-violation? (condition-predicate &syntax))
	 (define syntax-violation-form (condition-accessor &syntax (rtd-accessor &syntax 'form)))
	 (define syntax-violation-subform (condition-accessor &syntax (rtd-accessor &syntax 'subform)))

)
